from ..markup import Markup, get_installed_markups, get_markup_choices
from ..elements.elementbase import Attribute
from ..tags.content import RenderBase
from ..tags.context import DataSetter, LogicElement
from ..compat import text_type

from textwrap import dedent


class _Markup(RenderBase):
    """Insert markup in to content"""

    class Help:
        synopsis = "insert markup in to content"

    type = Attribute("Markup type", required=False, default="bbcode", choices=get_installed_markups())
    source = Attribute("Markup source", required=False, default=None, type="expression")

    def logic(self, context):
        type = self.type(context)
        if not Markup.supports(type):
            self.throw('markup.unsupported', "markup type '{}' is not supported".format(type))
        options = self.get_let_map(context)
        source_text = self.source(context) if self.has_parameter('source') else self.text
        text = self.source(context) or Markup.sub(type, context, source_text, options)
        markup = Markup(text, type, options)
        context['.content'].add_renderable(self._tag_name, markup)


class MarkupTag(RenderBase):

    source = Attribute("Markup source", required=False, default=None, type="expression")
    dedent = Attribute("De-dent source (remove common leading whitespace)?", type="boolean", default=True)

    class Help:
        undocumented = True

    def logic(self, context):
        if not Markup.supports(self.markup):
            self.throw('markup.unsupported', "markup type '{}' is not supported".format(self.markup))
        options = self.get_let_map(context)
        source_text = self.source(context) if self.has_parameter('source') else self.text
        if not self.has_parameter('source') and self.dedent(context):
            source_text = dedent(source_text)
        sub_text = Markup.sub(self.markup, context, source_text, options)
        markup = Markup(sub_text, self.markup, options)
        context['.content'].add_renderable(self._tag_name, markup)


class ProcessMarkup(DataSetter):
    """Process a given markup in to text"""

    type = Attribute("Markup type", required=False, default="bbcode")
    src = Attribute("Markup source", required=False, default=None, type="expression")
    dst = Attribute("Destination", type="reference", default=None)

    class Help:
        synopsis = "markup text"

    def logic(self, context):
        type = self.type(context)
        if not Markup.supports(type):
            self.throw('markup.unsupported', "markup type '{}' is not supported".format(type))
        options = self.get_let_map(context)
        source_text = self.src(context) if self.has_parameter('src') else self.text
        if not isinstance(source_text, text_type):
            self.throw('bad-value.unsupported-type',
                       "the 'src' parameter should be a a string (not {})".format(context.to_expr(source_text)))

        text = self.src(context) or Markup.sub(type, context, source_text, options)
        markup = Markup(text, type, options)
        result = markup.process(self.archive, context)
        self.set_context(context, self.dst(context), result)


class BBCode(MarkupTag):
    """Add bbcode to content"""
    markup = "bbcode"

    class Help:
        synopsis = "add bbcode to content"


class Markdown(MarkupTag):
    """Add markdown to content"""
    markup = "markdown"

    class Help:
        synopsis = "add markdown to content"


class GetMarkupTypes(DataSetter):
    """Get a list of all available Markup processors"""

    class Help:
        synopsis = "get supported markups"

    def get_value(self, context):
        return get_installed_markups()


class GetMarkupChoices(DataSetter):
    """Get a list of Markup processor choices, suitable for use in a [tag forms]select[/tag] tag."""

    class Help:
        synopsis = "get supported markups"

    def get_value(self, context):
        return get_markup_choices()


class MarkupInsert(LogicElement):
    """A callable invoked from Moya markup"""
    _moya_markup_insert = True

    class Help:
        synopsis = "insert code from markup"

    class Meta:
        is_call = True



from __future__ import unicode_literals
from __future__ import print_function

from ..cache import Cache

from collections import OrderedDict, namedtuple
from time import time as get_time
import logging


log = logging.getLogger('moya.runtime')


CacheEntry = namedtuple("CacheEntry", "value,expire_time")


class MemoryCache(Cache):
    """Caches in a memcached server"""
    cache_backend_name = "memory"

    def __init__(self, name, namespace, compress=True, compress_min=1024, size=1024*1024):
        super(MemoryCache, self).__init__(
            name,
            namespace,
            compress=compress,
            thread_safe=False)
        self.max_size = size
        self.entries = OrderedDict()
        self.size = 0

    @classmethod
    def initialize(cls, name, settings):
        return cls(
            name,
            settings.get('namespace', ''),
            compress=settings.get_bool('compress', True),
            compress_min=settings.get_int("compress_min", 16 * 1024),
            size=settings.get_int('size', 1024) * 1024,
        )

    def evict_entry(self, key):
        """Evict a single key."""
        if key in self.entries:
            entry = self.entries.pop(key)
            self.size -= entry.value

    def reclaim(self, num_bytes):
        """Reclaim at least `num_bytes`"""
        log.debug('%r size=%s bytes', self, self.size)
        log.debug('%r reclaiming %s bytes', self, num_bytes)
        reclaimed = 0
        while self.entries and reclaimed < num_bytes:
            key, entry = self.entries.popitem(last=False)
            log.debug("%r evicting %r", self, key)
            deleted_bytes_count = len(entry.value)
            self.size -= deleted_bytes_count
            reclaimed += deleted_bytes_count
        return reclaimed >= num_bytes

    def _get(self, key, default):
        try:
            value_bytes = self.entries[key].value
        except KeyError:
            return default

        # Remove entry from cache
        entry = self.entries.pop(key)
        self.size -= len(entry.value)

        # If it has expired return the default
        if entry.expire_time and get_time() > entry.expire_time:
            return default

        # Otherwise put it back in to the cache at first position
        self.entries[key] = entry
        self.size += len(entry.value)

        return self.decode_value(value_bytes)

    def _set(self, key, value, time):
        value_bytes = self.encode_value(value)
        value_size = len(value_bytes)
        if value_size > self.max_size:
            return
        self.evict_entry(key)
        if self.size + value_size > self.max_size:
            if not self.reclaim(self.size + value_size - self.max_size):
                return
        expire_time = None if time is None else get_time() + time / 1000.0
        self.entries[key] = CacheEntry(value_bytes, expire_time)
        self.size += value_size

    def _delete(self, key):
        if key in self.entries:
            self.size -= len(self.entries.pop(key).value)
